//========================================================================
// This program is distributed under the GPL license.
// Copyright 2004 Juha Nieminen
//========================================================================

#include "BWBitmap.hh"

#include <png.h>
#include <cstdio>
#include <csetjmp>
#include <cmath>


// Constructor and destructor
// --------------------------
BWBitmap::BWBitmap(unsigned width, unsigned height, unsigned char bgColor):
    bitmap_((width and height) ? new unsigned char[width*height] : 0),
    width_(width), height_(height),
    lineWidth_(.5), drawingColor_(0),
    font_(0), charWidth_(0), charHeight_(0)
{
    for(unsigned y=0; y<height; ++y)
        for(unsigned x=0; x<width; ++x)
            bitmap_[x+y*width] = bgColor;
}

BWBitmap::~BWBitmap()
{
    delete[] bitmap_;
}


// Raw pixel set and get
// ---------------------
inline void BWBitmap::put(unsigned x, unsigned y, unsigned char value)
{
    if(x < width_ and y < height_)
        bitmap_[x+y*width_] = value;
}

inline unsigned char BWBitmap::get(unsigned x, unsigned y) const
{
    return (x < width_ and y < height_) ? bitmap_[x+y*width_] : 0;
}


// Configuring methods
// -------------------
void BWBitmap::setLineWidth(double width)
{
    lineWidth_ = width/2;
}

void BWBitmap::setFont(const BWBitmap& fontBitmap,
                       unsigned charWidth, unsigned charHeight)
{
    font_ = &fontBitmap;
    charWidth_ = charWidth;
    charHeight_ = charHeight;
}


// Drawing methods
// ---------------
namespace
{
    // Calculates the partition of a pixel and a line (vx must be positive).
    // Returns how much the lower part of the line covers the pixel (0-1).
    double pixelPartition(double px, double py,
                          double x, double y, double vx, double vy)
    {
        double px1 = px-.5, px2 = px+.5, py1 = py-.5, py2 = py+.5;

        // If the line is completely horizontal:
        if(fabs(vy) < 1e-10)
        {
            if(y < py1) return 0;
            if(y > py2) return 1;
            return y-py1;
        }

        double t1 = (py1-y)/vy, t2 = (py2-y)/vy;
        double interX1 = x+t1*vx, interX2 = x+t2*vx;

        // Quick tests (line covers completely or doesn't touch):
        if((vy > 0 and interX2 < px1) or (vy < 0 and interX2 > px2)) return 1;
        if((vy > 0 and interX1 > px2) or (vy < 0 and interX1 < px1)) return 0;

        // If line crosses both vertical lines of the pixel:
        if((vy > 0 and interX1 < px1 and interX2 > px2) or
           (vy < 0 and interX2 < px1 and interX1 > px2))
        {
            t1 = (px1-x)/vx, t2 = (px2-x)/vx;
            double interY1 = y+t1*vy, interY2 = y+t2*vy;
            return (interY1+interY2)/2-py1;
        }

        // If line crosses both horizontal lines of the pixel:
        if(interX1 >= px1 and interX1 <= px2 and
           interX2 >= px1 and interX2 <= px2)
        {
            if(vy < 0)
                return (interX1+interX2)/2-px1;
            else
                return 1-((interX1+interX2)/2-px1);
        }

        // The line cuts the pixel into a triangle. Calculate its area:
        if(interX1 >= px1 and interX1 <= px2)
        {
            if(vy < 0) // lower-left triangle
            {
                t1 = (px1-x)/vx;
                double interY1 = y+t1*vy;
                return (interX1-px1)*(interY1-py1)/2;
            }
            else // lower-right triangle
            {
                t2 = (px2-x)/vx;
                double interY2 = y+t2*vy;
                return (px2-interX1)*(interY2-py1)/2;
            }
        }
        else
        {
            if(vy < 0) // upper-right triangle
            {
                t2 = (px2-x)/vx;
                double interY2 = y+t2*vy;
                return 1-(px2-interX2)*(py2-interY2)/2;
            }
            else // upper-left triangle
            {
                t1 = (px1-x)/vx;
                double interY1 = y+t1*vy;
                return 1-(interX2-px1)*(py2-interY1)/2;
            }
        }

        return 0;
    }

    // Calculates how much the given line covers the given pixel.
    // 0 = Does not touch the pixel. 1 = Covers it completely.
    // Line is defined as (x,y)+t*(vx,vy)
    // (vx,vy) must be a unit vector.
    double lineCoverage(int px, int py,
                        double x, double y, double vx, double vy,
                        double lineWidth)
    {
        if(vx < 0) { vx = -vx; vy = -vy; }
        if(vx < 1e-10) vx += 1e-9;
        double
            l1x = x-vy*lineWidth,
            l1y = y+vx*lineWidth,
            l2x = x+vy*lineWidth,
            l2y = y-vx*lineWidth;

        double upperLineCoverage = pixelPartition(px, py, l1x, l1y, vx, vy);
        double lowerLineCoverage = pixelPartition(px, py, l2x, l2y, vx, vy);

        return upperLineCoverage-lowerLineCoverage;
    }
}

void BWBitmap::line(double x1, double y1, double x2, double y2)
{
    if(y1 > y2)
    {
        double tmp = y1; y1 = y2; y2 = tmp;
        tmp = x1; x1 = x2; x2 = tmp;
    }

    double vx = x2-x1, vy = y2-y1;
    double len = std::sqrt(vx*vx+vy*vy);
    vx /= len; vy /= len;

    x1 += vx*lineWidth_; y1 += vy*lineWidth_;
    x2 -= vx*lineWidth_; y2 -= vy*lineWidth_;

    const int sy = int(y1-lineWidth_+.5), ey = int(y2+lineWidth_+.5);
    const int sx = (x1<x2) ? int(x1-lineWidth_+.5) : int(x2-lineWidth_+.5);
    const int ex = (x1<x2) ? int(x2+lineWidth_+.5) : int(x1+lineWidth_+.5);

    for(int py=sy; py<=ey; ++py)
        for(int px=sx; px<=ex; ++px)
        {
            double value = lineCoverage(px,py, x1,y1, vx,vy, lineWidth_);
            put(px,py, (unsigned char)
                (drawingColor_*value + get(px,py)*(1-value) + .5));
        }
}

void BWBitmap::circle(double cx, double cy, double radius)
{
    const int sx = int(cx-radius-lineWidth_), ex = int(cx+radius+lineWidth_+1);
    const int sy = int(cy-radius-lineWidth_), ey = int(cy+radius+lineWidth_+1);

    for(int py = sy; py <= ey; ++py)
        for(int px = sx; px <= ex; ++px)
        {
            double x = px - cx + 1e-5, y = py - cy;
            double len = std::sqrt(x*x+y*y);
            x /= len;
            y /= len;
            double vx = -y, vy = x;
            x = cx + radius*x;
            y = cy + radius*y;

            double value = lineCoverage(px, py, x, y, vx, vy, lineWidth_);
            put(px, py, (unsigned char)
                (drawingColor_*value + get(px,py)*(1-value) + .5));
        }
}

void BWBitmap::filledCircle(double cx, double cy, double radius)
{
    const int sx = int(cx-radius-lineWidth_), ex = int(cx+radius+lineWidth_+1);
    const int sy = int(cy-radius-lineWidth_), ey = int(cy+radius+lineWidth_+1);

    for(int py = sy; py <= ey; ++py)
        for(int px = sx; px <= ex; ++px)
        {
            double x = px - cx + 1e-5, y = py - cy;
            double len = std::sqrt(x*x+y*y);
            x /= len;
            y /= len;
            double vx = -y, vy = x;
            x = cx + (radius+lineWidth_)*x;
            y = cy + (radius+lineWidth_)*y;

            if(vx < 0) { vx = -vx; vy = -vy; }
            if(vx < 1e-10) vx += 1e-9;

            double value = pixelPartition(px, py, x, y, vx, vy);
            if(py > cy)
                put(px, py, (unsigned char)
                    (drawingColor_*value + get(px,py)*(1-value) + .5));
            else
                put(px, py, (unsigned char)
                    (drawingColor_*(1-value) + get(px,py)*value + .5));
        }
}

void BWBitmap::character(unsigned charIndex, unsigned x, unsigned y)
{
    if(!font_) return;

    const unsigned fontWidth = font_->width_/charWidth_;
    const unsigned cx = (charIndex%fontWidth)*charWidth_;
    const unsigned cy = (charIndex/fontWidth)*charHeight_;
    const unsigned bmx = x-charWidth_/2;
    const unsigned bmy = y-charHeight_/2;

    for(unsigned iy = 0; iy < charHeight_; ++iy)
        for(unsigned ix = 0; ix < charWidth_; ++ix)
        {
            double value = font_->get(cx+ix, cy+iy)/255.0;
            put(bmx+ix, bmy+iy, (unsigned char)
                (drawingColor_*(1-value) + get(bmx+ix, bmy+iy)*value + .5));
        }
}



// PNG reading and writing
// -----------------------
namespace
{
    struct FilePtr
    {
        std::FILE* fp;

        FilePtr(): fp(0) {}
        ~FilePtr() { if(fp) std::fclose(fp); }
    };

    struct PNGStructPtrs
    {
        png_structp read;
        png_structp write;
        png_infop info;

        PNGStructPtrs(): read(0), write(0), info(0) {}
        ~PNGStructPtrs()
        {
            if(read)
                if(info) png_destroy_read_struct(&read, &info, 0);
                else png_destroy_read_struct(&read, 0, 0);
            else if(write)
                if(info) png_destroy_write_struct(&write, &info);
                else png_destroy_write_struct(&write, 0);
        }
    };
}

bool BWBitmap::readPNG(const std::string& filename)
{
    FilePtr iFile;
    iFile.fp = std::fopen(filename.c_str(), "rb");
    if(!iFile.fp)
    {
        std::fprintf(stderr, "Can't open ");
        std::perror(filename.c_str());
        return false;
    }

    png_byte header[8];
    std::fread(header, 1, 8, iFile.fp);
    if(png_sig_cmp(header, 0, 8))
    {
        std::fprintf(stderr, "%s is not a PNG file.\n", filename.c_str());
        return false;
    }
    std::fseek(iFile.fp, 0, SEEK_SET);

    PNGStructPtrs ptrs;
    ptrs.read = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    if(!ptrs.read) return false;

    ptrs.info = png_create_info_struct(ptrs.read);
    if(!ptrs.info) return false;

    if(setjmp(png_jmpbuf(ptrs.read))) return false;

    png_init_io(ptrs.read, iFile.fp);
    png_read_info(ptrs.read, ptrs.info);

    if(ptrs.info->color_type != PNG_COLOR_TYPE_GRAY or
       ptrs.info->bit_depth != 8)
    {
        fprintf(stderr, "%s is not a gray-scale 8-bit PNG.\n",
                filename.c_str());
        return false;
    }

    delete[] bitmap_;

    width_ = ptrs.info->width;
    height_ = ptrs.info->height;

    bitmap_ = new unsigned char[width_*height_];

    png_read_update_info(ptrs.read, ptrs.info);

    unsigned char** rowPointers = new unsigned char*[height_];
    for(unsigned i=0; i<height_; ++i)
        rowPointers[i] = &bitmap_[i*width_];

    png_read_image(ptrs.read, rowPointers);

    delete[] rowPointers;
    return true;
}

bool BWBitmap::writePNG(const std::string& filename)
{
    FilePtr oFile;
    oFile.fp = std::fopen(filename.c_str(), "wb");
    if(!oFile.fp)
    {
        std::fprintf(stderr, "Can't create ");
        std::perror(filename.c_str());
        return false;
    }

    PNGStructPtrs ptrs;
    ptrs.write = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
    if(!ptrs.write) return false;

    ptrs.info = png_create_info_struct(ptrs.write);
    if(!ptrs.info) return false;

    if(setjmp(png_jmpbuf(ptrs.write))) return false;

    png_init_io(ptrs.write, oFile.fp);

    png_set_IHDR(ptrs.write, ptrs.info, width_, height_, 8,
                 PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    png_write_info(ptrs.write, ptrs.info);

    unsigned char** rowPointers = new unsigned char*[height_];
    for(unsigned i=0; i<height_; ++i)
        rowPointers[i] = &bitmap_[i*width_];

    png_write_image(ptrs.write, rowPointers);
    png_write_end(ptrs.write, 0);

    delete[] rowPointers;

    return true;
}
